import pandas as pd

from .auth import AuthClient
from .gcs import GCSFileSystem


class FileClient:
    @staticmethod
    def _ensure_gcs_uri_prefix(gcs_path):
        """
        GCS uri is the gcs file path prefixed with 'gs://'. Some operations require GCS uris,
        but we don't want the user to bother with knowing when to use the prefix,
        so we ensure its presence automatically where it is needed
        :param gcs_path:
        :return:
        """
        gs_uri_prefix = "gs://"
        if not gcs_path.startswith(gs_uri_prefix):
            gcs_path = f"{gs_uri_prefix}{gcs_path}"
        return gcs_path

    @staticmethod
    def get_gcs_file_system(**kwargs):
        """
        Return a pythonic file-system for Google Cloud Storage - initialized with a personal Google Identity token.
        See https://gcsfs.readthedocs.io/en/latest for usage
        """
        return GCSFileSystem(token=AuthClient.fetch_google_credentials(), **kwargs)

    @staticmethod
    def ls(gcs_path, detail=False, **kwargs):
        """
        List the contents of a GCS bucket path
        :param gcs_path: GCS bucket path
        :param detail: show file details
        :return: Array of contents of bucket
        """
        return FileClient.get_gcs_file_system().ls(gcs_path, detail=detail, **kwargs)

    @staticmethod
    def cat(gcs_path):
        """
        Get string content of file from gcs
        :param gcs_path: path or paths to the file(s) you want to get the contents of
        :return: utf-8 decoded string content of the given file
        """
        return FileClient.get_gcs_file_system().cat(gcs_path).decode("utf-8")

    @staticmethod
    def gcs_open(gcs_path, mode='r'):
        """
        Open a file in GCS, works like regular python open()
        :param gcs_path:
        :param mode:
        :return:
        """
        FileClient.get_gcs_file_system().open(FileClient._ensure_gcs_uri_prefix(gcs_path), mode)

    @staticmethod
    def load_csv_to_pandas(gcs_path):
        """
        Reads a csv file from google cloud storage into a Pandas data frame
        :param gcs_path: of the file, starting with the bucket name
        :return: a Pandas data frame
        """
        return pd.read_csv(FileClient._ensure_gcs_uri_prefix(gcs_path), storage_options={"token": AuthClient.fetch_google_credentials()})

    @staticmethod
    def load_json_to_pandas(gcs_path):
        """
        Reads a json file from google cloud storage into a Pandas data frame
        :param gcs_path: of the file, starting with the bucket name
        :return: a Pandas data frame
        """
        return pd.read_json(FileClient._ensure_gcs_uri_prefix(gcs_path), storage_options={"token": AuthClient.fetch_google_credentials()})

    @staticmethod
    def load_xml_to_pandas(gcs_path):
        """
        Reads an xml file from google cloud storage into a Pandas data frame
        :param gcs_path: of the file, starting with the bucket name
        :return: a Pandas data frame
        """
        return pd.read_xml(FileClient._ensure_gcs_uri_prefix(gcs_path), storage_options={"token": AuthClient.fetch_google_credentials()})

    @staticmethod
    def save_pandas_to_csv(df: pd.DataFrame, gcs_path, index=False):
        """
        Write the contents of a Pandas data frame to a csv file in a bucket
        :param df: the Pandas data frame to persist as csv
        :param gcs_path: target path, starting with the bucket name and ending with the file name
        :param index: True if you want to write the pandas index to the file
        :return:
        """
        df.to_csv(FileClient._ensure_gcs_uri_prefix(gcs_path), storage_options={"token": AuthClient.fetch_google_credentials()}, index=index)

    @staticmethod
    def save_pandas_to_json(df: pd.DataFrame, gcs_path):
        """
        Write the contents of a Pandas data frame to a json file in a bucket
        :param df: the Pandas data frame to persist as json
        :param gcs_path: target path, starting with the bucket name and ending with the file name
        :param index: True if you want to write the pandas index to the file
        :return:
        """
        df.to_json(FileClient._ensure_gcs_uri_prefix(gcs_path), storage_options={"token": AuthClient.fetch_google_credentials()})

    @staticmethod
    def save_pandas_to_xml(df: pd.DataFrame, gcs_path, index=False):
        """
        Write the contents of a Pandas data frame to an xml file in a bucket
        :param df: the Pandas data frame to persist as xml
        :param gcs_path: target path, starting with the bucket name and ending with the file name
        :param index: True if you want to write the pandas index to the file
        :return:
        """
        df.to_xml(FileClient._ensure_gcs_uri_prefix(gcs_path), storage_options={"token": AuthClient.fetch_google_credentials()}, index=index)



import json

import requests

from .auth import AuthClient


class CollectorClient:

    def __init__(self, collector_url):
        self.collector_url = collector_url

    def initiate(self, specification):
        keycloak_token = AuthClient.fetch_personal_token()
        collector_response = requests.put(
            self.collector_url,
            headers={'Authorization': 'Bearer %s' % keycloak_token, 'Content-type': 'application/json'},
            data=json.dumps(specification)
        )

        if collector_response.status_code == 201:
            print("Collector initiated successfully! Check running tasks with the list_tasks() method.")
        else:
            raise RuntimeError(f'Something went wrong. Response from {self.collector_url}:'
                               f' {collector_response.status_code}'
                               f' - {collector_response.text or collector_response.reason}')

    def list_tasks(self):
        keycloak_token = AuthClient.fetch_personal_token()
        collector_response = requests.get(self.collector_url, headers={'Authorization': 'Bearer %s' % keycloak_token})

        if collector_response.status_code == 200:
            print(json.dumps(collector_response.json(), indent=2))
        else:
            raise RuntimeError(f'Something went wrong. Response from {self.collector_url}:'
                               f' {collector_response.status_code}'
                               f' - {collector_response.text or collector_response.reason}')



__version__ = "1.2.1"

from .auth import AuthClient
from .files import FileClient
from .guardian import GuardianClient
from .collector import CollectorClient
from .backports import show, details, read_pandas, write_pandas



from .files import FileClient
from .auth import AuthClient
from pandas import read_csv, read_json
# import utils to register the pyarrow extension types
import pandas.core.arrays._arrow_utils  # noqa


def show(gcs_path):
    """
    Backported dapla function to recursively show all folders below a given GCS path
    :param gcs_path: the path from which you want to list all folders
    :return: a simplified list of files or folders
    """
    fs = FileClient.get_gcs_file_system()
    out = dict()
    files = None
    for path, dirs, files in fs.walk(gcs_path, detail=True):
        out.update({_trimmed_name(info): info for name, info in dirs.items()})
    # Add the base path (if it exists) to avoid an empty list when there are no subfolders
    if len(out) == 0 and files is not None:
        # Get the bucket name from any of the files
        bucket_name = list(files.values())[0]['bucket']
        trimmed_name = gcs_path.lstrip('gs://').lstrip(bucket_name).rstrip('/')
        out[trimmed_name] = {}
    return sorted(out)


def details(gcs_path):
    """
    Backported dapla function to support detailed list of files for a given GCS path
    :param gcs_path: path to the file(s) you want to get the contents of
    :return: a detailed list of files
    """
    fs = FileClient.get_gcs_file_system()
    return list(map(lambda o: _folder_item(o) if o['storageClass'] == 'DIRECTORY' else _file_item(o),
                    fs.ls(gcs_path, detail=True)))


def read_pandas(gcs_path, file_format="parquet", columns=None, **kwargs):
    """
    Backported dapla function read a dataset for a given GCS path and convert it to a Pandas dataframe.
    :param gcs_path: path to the directory or file you want to get the contents of
    :param file_format: the expected file format. All file formats other than "parquet" are delegated to Pandas'
        read_table method
    :param columns: If not None, only these columns will be read
    :return: a pandas dataframe

    Other arguments are passed through to the to_pandas method.
    """
    if file_format == "parquet":
        import pyarrow.parquet as pq
        fs = FileClient.get_gcs_file_system()
        parquet_ds = pq.ParquetDataset(gcs_path, filesystem=fs)
        return parquet_ds.read_pandas(columns=columns).to_pandas(split_blocks=False, self_destruct=True, **kwargs)
    elif file_format == "json":
        return read_json(gcs_path, storage_options=_get_storage_options(), **kwargs)
    elif file_format == "csv":
        return read_csv(gcs_path, storage_options=_get_storage_options(), **kwargs)
    else:
        raise ValueError(f"Invalid file format {file_format}")


def write_pandas(df, gcs_path, file_format="parquet", **kwargs):
    """
    Backported dapla function to write a pandas dataset to a given GCS path.
    :param df: the pandas dataset
    :param gcs_path: path to the file you want to write to - must have a suffix that corresponds to the file_format
    :param file_format: file format of file you want to write to
    Other arguments are passed through to the pyarrow write_table method.
    """
    import pyarrow.parquet
    import pandas.io.parquet
    pandas.io.parquet.BaseImpl.validate_dataframe(df)
    fs = FileClient.get_gcs_file_system()

    if file_format == "parquet":
        # Transfom and write pandas dataframe
        from_pandas_kwargs = {"schema": kwargs.pop("schema", None)}
        table = pyarrow.Table.from_pandas(df, preserve_index=True, **from_pandas_kwargs)
        if '.parquet' not in gcs_path:
            raise Exception('Path must be a parquet file')
        with fs.open(gcs_path, mode="wb") as buffer:
            pyarrow.parquet.write_table(
                table,
                buffer,
                flavor={'spark'},
                compression="snappy",
                coerce_timestamps="ms",
                **kwargs,
            )
    elif file_format == "json":
        df.to_json(gcs_path, storage_options=_get_storage_options(), **kwargs)
    elif file_format == "csv":
        df.to_csv(gcs_path, storage_options=_get_storage_options(), **kwargs)
    else:
        raise ValueError(f"Invalid file format {file_format}")


def _get_storage_options():
    token = AuthClient.fetch_google_credentials()
    return {"token": token} if token is not None else None


def _trimmed_name(o):
    return o['name'].lstrip(o['bucket'])


def _folder_item(o):
    return {'Size': o['size'], 'Name': _trimmed_name(o)}


def _file_item(o):
    return {'Size': o['size'], 'Name': _trimmed_name(o), 'Created': o['timeCreated'], 'Updated': o['updated']}
